local ADDON = ...

local Core = CogWheel("LibModule"):GetModule(ADDON)
if (not Core) then 
	return 
end

local Module = Core:NewModule("UnitFrameTarget", "LibEvent", "LibUnitFrame", "LibSound")
local Colors = CogWheel("LibDB"):GetDatabase(ADDON..": Colors")
local Auras = CogWheel("LibDB"):GetDatabase(ADDON..": Auras")
local Fonts = CogWheel("LibDB"):GetDatabase(ADDON..": Fonts")
local Layout = CogWheel("LibDB"):GetDatabase(ADDON..": Layout [UnitFrameTarget]")

-- Lua API
local _G = _G
local unpack = unpack

-- WoW API
local GetAccountExpansionLevel = _G.GetAccountExpansionLevel
local GetExpansionLevel = _G.GetExpansionLevel
local GetQuestGreenRange = _G.GetQuestGreenRange
local IsXPUserDisabled = _G.IsXPUserDisabled
local UnitClassification = _G.UnitClassification
local UnitExists = _G.UnitExists
local UnitIsConnected = _G.UnitIsConnected
local UnitIsDeadOrGhost = _G.UnitIsDeadOrGhost
local UnitIsEnemy = _G.UnitIsEnemy
local UnitIsFriend = _G.UnitIsFriend
local UnitIsTrivial = _G.UnitIsTrivial
local UnitLevel = _G.UnitLevel

-- WoW Constants & Objects
local DEAD = _G.DEAD
local MAX_PLAYER_LEVEL_TABLE = _G.MAX_PLAYER_LEVEL_TABLE
local PLAYER_OFFLINE = _G.PLAYER_OFFLINE

-- Player Class
local _, PlayerClass = UnitClass("player")

-- Current player level
-- We use this to decide how dangerous enemies are 
-- relative to our current character.
local LEVEL = UnitLevel("player") 

-- Constants to hold various info about our last target 
-- We need this to decide when the artwork should change
local TARGET_STYLE


-- Utility Functions
-----------------------------------------------------------------

-- Returns the correct difficulty color compared to the player
local getDifficultyColorByLevel = function(level)
	level = level - LEVEL
	if (level > 4) then
		return Colors.quest.red.colorCode
	elseif (level > 2) then
		return Colors.quest.orange.colorCode
	elseif (level >= -2) then
		return Colors.quest.yellow.colorCode
	elseif (level >= -GetQuestGreenRange()) then
		return Colors.quest.green.colorCode
	else
		return Colors.quest.gray.colorCode
	end
end

-- Proxy function to get media from our local media folder
local getPath = function(fileName)
	return ([[Interface\AddOns\%s\media\%s.tga]]):format(ADDON, fileName)
end 

-- Figure out if the player has a XP bar
local PlayerHasXP = function()
	local playerLevel = UnitLevel("player")
	local expacMax = MAX_PLAYER_LEVEL_TABLE[LE_EXPANSION_LEVEL_CURRENT or #MAX_PLAYER_LEVEL_TABLE]
	local playerMax = MAX_PLAYER_LEVEL_TABLE[GetAccountExpansionLevel() or #MAX_PLAYER_LEVEL_TABLE]
	local hasXP = (not IsXPUserDisabled()) and ((playerLevel < playerMax) or (playerLevel < expacMax))
	return hasXP
end


-- Callbacks
-----------------------------------------------------------------

-- Number abbreviations
local OverrideValue = function(element, unit, min, max, disconnected, dead, tapped)
	if (min >= 1e8) then 		element.Value:SetFormattedText("%dm", min/1e6) 		-- 100m, 1000m, 2300m, etc
	elseif (min >= 1e6) then 	element.Value:SetFormattedText("%.1fm", min/1e6) 	-- 1.0m - 99.9m 
	elseif (min >= 1e5) then 	element.Value:SetFormattedText("%dk", min/1e3) 		-- 100k - 999k
	elseif (min >= 1e3) then 	element.Value:SetFormattedText("%.1fk", min/1e3) 	-- 1.0k - 99.9k
	elseif (min > 0) then 		element.Value:SetText(min) 							-- 1 - 999
	else 						element.Value:SetText("")
	end 
end 

local OverrideHealthValue = function(element, unit, min, max, disconnected, dead, tapped)
	if disconnected then 
		if element.Percent then 
			element.Percent:SetText("")
		end 
		if element.Value then 
			element.Value:SetText(PLAYER_OFFLINE)
		end 
	elseif dead then 
		if element.Percent then 
			element.Percent:SetText("")
		end 
		if element.Value then 
			element.Value:SetText(DEAD)
		end 
	else
		if element.Percent then 
			element.Percent:SetFormattedText("%d", min/max*100 - (min/max*100)%1)
		end 
		if element.Value then 
			OverrideValue(element, unit, min, max, disconnected, dead, tapped)
		end 
	end 
end 

local Threat_UpdateColor = function(element, unit, status, r, g, b)
	if element.health then 
		element.health:SetVertexColor(r, g, b)
	end
	if element.power then 
		element.power:SetVertexColor(r, g, b)
	end
	if element.mana then 
		element.mana:SetVertexColor(r, g, b)
	end 
end

local Threat_IsShown = function(element)
	return element.health and element.health:IsShown()
end 

local Threat_Show = function(element)
	if 	element.health then 
		element.health:Show()
	end
	if 	element.power then 
		element.power:Show()
	end
	if element.mana then 
		element.mana:Show()
	end 
end 

local Threat_Hide = function(element)
	if 	element.health then 
		element.health:Hide()
	end 
	if element.power then 
		element.power:Hide()
	end
	if element.mana then 
		element.mana:Hide()
	end
end 

local PostCreateAuraButton = function(element, button)
	
	button.Icon:SetTexCoord(unpack(Layout.AuraIconTexCoord))
	button.Icon:SetSize(unpack(Layout.AuraIconSize))
	button.Icon:ClearAllPoints()
	button.Icon:SetPoint(unpack(Layout.AuraIconPlace))

	button.Count:SetFontObject(Layout.AuraCountFont)
	button.Count:ClearAllPoints()
	button.Count:SetPoint(unpack(Layout.AuraCountPlace))

	button.Time:SetFontObject(Layout.AuraTimeFont)
	button.Time:ClearAllPoints()
	button.Time:SetPoint(unpack(Layout.AuraTimePlace))

	local layer, level = button.Icon:GetDrawLayer()

	button.Darken = button:CreateTexture()
	button.Darken:SetDrawLayer(layer, level + 1)
	button.Darken:SetSize(button.Icon:GetSize())
	button.Darken:SetAllPoints(button.Icon)
	button.Darken:SetColorTexture(0, 0, 0, .25)

	button.Border = button.Overlay:CreateFrame("Frame")
	button.Border:Place(unpack(Layout.AuraBorderFramePlace))
	button.Border:SetSize(unpack(Layout.AuraBorderFrameSize))
	button.Border:SetBackdrop(Layout.AuraBorderBackdrop)
	button.Border:SetBackdropColor(unpack(Layout.AuraBorderBackdropColor))
	button.Border:SetBackdropBorderColor(unpack(Layout.AuraBorderBackdropBorderColor))

end

local PostUpdateAuraButton = function(element, button)
	if (not button) or (not button:IsVisible()) then 
		return 
	end 
end

-- not needed in gui, needed in azerite. fix later. 
local PostUpdateTextures = function(self)
	if (not Layout.UseProgressiveFrames) then 
		return
	end 

	-- Figure out if the various artwork and bar textures need to be updated
	-- We could put this into element post updates, 
	-- but to avoid needless checks we limit this to actual target updates. 
	local unitLevel = UnitLevel("target")
	local unitClassification = UnitClassification("target")

	if ((unitClassification == "worldboss") or (unitLevel and (unitLevel < 1))) then 
		if (TARGET_STYLE ~= "BOSS") then 
			TARGET_STYLE = "BOSS"

			self.Health:SetSize(unpack(Layout.BossHealthSize))
			self.Health:Place(unpack(Layout.BossHealthPlace))
			self.Health:SetStatusBarTexture(Layout.BossHealthTexture)
			self.Health:SetSparkMap(Layout.BossHealthSparkMap)
	
			if Layout.UseHealthBackdrop then 
				self.Health.Bg:SetSize(unpack(Layout.BossHealthBackdropSize))
				self.Health.Bg:SetPoint(unpack(Layout.BossHealthBackdropPlace))
				self.Health.Bg:SetTexture(Layout.BossHealthBackdropTexture)
				self.Health.Bg:SetVertexColor(unpack(Layout.BossHealthBackdropColor))
			end 

			self.Health.Value:Show()
			self.Health.Percent:Show()

			self.Threat.health:SetSize(694, 190)
			self.Threat.health:SetPoint("CENTER", -.5, 1 +1)
			self.Threat.health:SetTexture(getPath("hp_boss_case_glow"))
	
			self.Cast:SetSize(533, 40)
			self.Cast:Place("TOPRIGHT", -27, -27)
			self.Cast:SetStatusBarTexture(getPath("hp_boss_bar"))
			self.Cast:SetSparkMap(map.boss)

			local portraitFg = self.Portrait.Fg
			portraitFg:SetTexture(getPath("portrait_frame_hi"))
			portraitFg:SetVertexColor(Colors.ui.stone[1], Colors.ui.stone[2], Colors.ui.stone[3])
		end

	-- War Seasoned / Capped  
	elseif (unitLevel >= MAX_PLAYER_LEVEL_TABLE[GetExpansionLevel()]) or (UnitIsUnit("target", "player") and (not PlayerHasXP())) then 
		if (TARGET_STYLE ~= "CAP") then 
			TARGET_STYLE = "CAP"

			local health = self.Health
			health:SetSize(385, 40)
			health:Place("TOPRIGHT", -27, -27)
			health:SetStatusBarTexture(getPath("hp_cap_bar"))
			health:SetSparkMap(map.normal)
	
			local healthBg = self.Health.Bg
			healthBg:SetSize(716, 188)
			healthBg:SetPoint("CENTER", -1, .5)
			healthBg:SetTexture(getPath("hp_cap_case"))
			healthBg:SetVertexColor(Colors.ui.stone[1], Colors.ui.stone[2], Colors.ui.stone[3])

			local healthVal = self.Health.Value
			healthVal:Show()

			local healthPerc = self.Health.Percent
			healthPerc:Hide()

			local threat = self.Threat
			threat.health:SetSize(716, 188)
			threat.health:SetPoint("CENTER", -1, .5  +1)
			threat.health:SetTexture(getPath("hp_cap_case_glow"))

			local cast = self.Cast
			cast:SetSize(385, 40)
			cast:Place("TOPRIGHT", -27, -27)
			cast:SetStatusBarTexture(getPath("hp_cap_bar"))
			cast:SetSparkMap(map.normal)

			local portraitFg = self.Portrait.Fg
			portraitFg:SetTexture(getPath("portrait_frame_hi"))
			portraitFg:SetVertexColor(Colors.ui.stone[1], Colors.ui.stone[2], Colors.ui.stone[3])
		end 

	-- Battle Hardened / Mid level
	elseif (unitLevel >= 40) then 
		if (TARGET_STYLE ~= "MID") then 
			TARGET_STYLE = "MID"

			local health = self.Health
			health:SetSize(385, 37)
			health:Place("TOPRIGHT", -27, -27)
			health:SetStatusBarTexture(getPath("hp_lowmid_bar"))
			health:SetSparkMap(map.normal)
	
			local healthBg = self.Health.Bg
			healthBg:SetSize(716, 188)
			healthBg:SetPoint("CENTER", -1, -.5)
			healthBg:SetTexture(getPath("hp_mid_case"))
			healthBg:SetVertexColor(Colors.ui.stone[1], Colors.ui.stone[2], Colors.ui.stone[3])
	
			local healthVal = self.Health.Value
			healthVal:Show()

			local healthPerc = self.Health.Percent
			healthPerc:Hide()

			local threat = self.Threat
			threat.health:SetSize(716, 188)
			threat.health:SetPoint("CENTER", -1, -.5 +1)
			threat.health:SetTexture(getPath("hp_mid_case_glow"))

			local cast = self.Cast
			cast:SetSize(385, 37)
			cast:Place("TOPRIGHT", -27, -27)
			cast:SetStatusBarTexture(getPath("hp_lowmid_bar"))
			cast:SetSparkMap(map.normal)

			local portraitFg = self.Portrait.Fg
			portraitFg:SetTexture(getPath("portrait_frame_hi"))
			portraitFg:SetVertexColor(Colors.ui.stone[1], Colors.ui.stone[2], Colors.ui.stone[3])
		end 

	-- Trivial / Critter
	elseif ((unitLevel == 1) and (not UnitIsPlayer("target"))) then 
		if (TARGET_STYLE ~= "CRITTER") then 
			TARGET_STYLE = "CRITTER"

			local health = self.Health
			health:SetSize(40, 36)
			health:Place("TOPRIGHT", -24, -24)
			health:SetStatusBarTexture(getPath("hp_critter_bar"))
			health:SetSparkMap(map.critter)
	
			local healthBg = self.Health.Bg
			healthBg:SetSize(98,96)
			healthBg:SetPoint("CENTER", 0, 1)
			healthBg:SetTexture(getPath("hp_critter_case"))
			healthBg:SetVertexColor(unpack(Colors.ui.wood))

			local healthVal = self.Health.Value
			healthVal:Hide()

			local healthPerc = self.Health.Percent
			healthPerc:Hide()

			local threat = self.Threat
			threat.health:SetSize(98,96)
			threat.health:SetPoint("CENTER", 0, 1 +1)
			threat.health:SetTexture(getPath("hp_critter_case_glow"))

			local cast = self.Cast
			cast:SetSize(40, 36)
			cast:Place("TOPRIGHT", -24, -24)
			cast:SetStatusBarTexture(getPath("hp_critter_bar"))
			cast:SetSparkMap(map.critter)

			local portraitFg = self.Portrait.Fg
			portraitFg:SetTexture(getPath("portrait_frame_lo"))
			portraitFg:SetVertexColor(unpack(Colors.ui.wood))
		end

		-- Novice / Low Level
	elseif (unitLevel > 1) or UnitIsPlayer("target") then 

		if (TARGET_STYLE ~= "LOW") then 
			TARGET_STYLE = "LOW" 

			local health = self.Health
			health:SetSize(385, 37)
			health:Place("TOPRIGHT", -27, -27)
			health:SetStatusBarTexture(getPath("hp_lowmid_bar"))
			health:SetSparkMap(map.normal)

			local healthVal = self.Health.Value
			healthVal:Show()

			local healthPerc = self.Health.Percent
			healthPerc:Hide()

			local healthBg = self.Health.Bg
			healthBg:SetSize(716, 188)
			healthBg:SetPoint("CENTER", -1, -.5)
			healthBg:SetTexture(getPath("hp_low_case"))
			healthBg:SetVertexColor(unpack(Colors.ui.wood))

			local threat = self.Threat
			threat.health:SetSize(716, 188)
			threat.health:SetPoint("CENTER", -1, -.5  +1)
			threat.health:SetTexture(getPath("hp_low_case_glow"))

			local cast = self.Cast
			cast:SetSize(385, 37)
			cast:Place("TOPRIGHT", -27, -27)
			cast:SetStatusBarTexture(getPath("hp_lowmid_bar"))
			cast:SetSparkMap(map.normal)

			local absorb = self.Absorb
			absorb:SetSize(385, 37)
			absorb:Place("TOPRIGHT", -27, -27)
			absorb:SetStatusBarTexture(getPath("hp_lowmid_bar"))
			absorb:SetSparkMap(map.normal)

			local portraitFg = self.Portrait.Fg
			portraitFg:SetTexture(getPath("portrait_frame_lo"))
			portraitFg:SetVertexColor(unpack(Colors.ui.wood)) 
		end 

	end 



	if (not PlayerHasXP()) then 
		self.Health:SetSize(unpack(Layout.SeasonedHealthSize))
		self.Health:SetStatusBarTexture(Layout.SeasonedHealthTexture)

		if Layout.UseHealthBackdrop then 
			self.Health.Bg:SetTexture(Layout.SeasonedHealthBackdropTexture)
			self.Health.Bg:SetVertexColor(unpack(Layout.SeasonedHealthBackdropColor))
		end 

		if Layout.UseThreat then
			if self.Threat.health then 
				self.Threat.health:SetTexture(Layout.SeasonedHealthThreatTexture)
			end 
		end

		if Layout.UsePowerBar then 
			if Layout.UsePowerForeground then 
				self.Power.Fg:SetTexture(Layout.SeasonedPowerForegroundTexture)
				self.Power.Fg:SetVertexColor(unpack(Layout.SeasonedPowerForegroundColor))
			end
		end

		if Layout.UseAbsorbBar then 
			self.Absorb:SetSize(unpack(Layout.SeasonedAbsorbSize))
			self.Absorb:SetStatusBarTexture(Layout.SeasonedAbsorbTexture)
		end 

		if Layout.UseCastBar then 
			self.Cast:SetSize(unpack(Layout.SeasonedCastSize))
			self.Cast:SetStatusBarTexture(Layout.SeasonedCastTexture)
		end

		if Layout.UseMana then 
			self.ExtraPower.Border:SetTexture(Layout.SeasonedManaOrbTexture)
			self.ExtraPower.Border:SetVertexColor(unpack(Layout.SeasonedManaOrbColor)) 
		end 

	elseif (LEVEL >= Layout.HardenedLevel) then 
		self.Health:SetSize(unpack(Layout.HardenedHealthSize))
		self.Health:SetStatusBarTexture(Layout.HardenedHealthTexture)

		if Layout.UseHealthBackdrop then 
			self.Health.Bg:SetTexture(Layout.HardenedHealthBackdropTexture)
			self.Health.Bg:SetVertexColor(unpack(Layout.HardenedHealthBackdropColor))
		end

		if Layout.UseThreat then
			if self.Threat.health then 
				self.Threat.health:SetTexture(Layout.HardenedHealthThreatTexture)
			end 
		end 

		if Layout.UsePowerBar then 
			if Layout.UsePowerForeground then 
				self.Power.Fg:SetTexture(Layout.HardenedPowerForegroundTexture)
				self.Power.Fg:SetVertexColor(unpack(Layout.HardenedPowerForegroundColor))
			end
		end

		if Layout.UseAbsorbBar then 
			self.Absorb:SetSize(unpack(Layout.HardenedAbsorbSize))
			self.Absorb:SetStatusBarTexture(Layout.HardenedAbsorbTexture)
		end

		if Layout.UseCastBar then 
			self.Cast:SetSize(unpack(Layout.HardenedCastSize))
			self.Cast:SetStatusBarTexture(Layout.HardenedCastTexture)
		end

		if Layout.UseMana then 
			self.ExtraPower.Border:SetTexture(Layout.HardenedManaOrbTexture)
			self.ExtraPower.Border:SetVertexColor(unpack(Layout.HardenedManaOrbColor)) 
		end 

	elseif ((unitLevel == 1) and (not UnitIsPlayer("target"))) then 


	else

		self.Health:SetSize(unpack(Layout.NoviceHealthSize))
		self.Health:SetStatusBarTexture(Layout.NoviceHealthTexture)

		if Layout.UseHealthBackdrop then 
			self.Health.Bg:SetTexture(Layout.NoviceHealthBackdropTexture)
			self.Health.Bg:SetVertexColor(unpack(Layout.NoviceHealthBackdropColor))
		end

		if Layout.UseThreat then
			if self.Threat.health then 
				self.Threat.health:SetTexture(Layout.NoviceHealthThreatTexture)
			end 
		end 

		if Layout.UsePowerBar then 
			if Layout.UsePowerForeground then 
				self.Power.Fg:SetTexture(Layout.NovicePowerForegroundTexture)
				self.Power.Fg:SetVertexColor(unpack(Layout.NovicePowerForegroundColor))
			end
		end

		if Layout.UseAbsorbBar then 
			self.Absorb:SetSize(unpack(Layout.NoviceAbsorbSize))
			self.Absorb:SetStatusBarTexture(Layout.NoviceAbsorbTexture)
		end

		if Layout.UseCastBar then 
			self.Cast:SetSize(unpack(Layout.NoviceCastSize))
			self.Cast:SetStatusBarTexture(Layout.NoviceCastTexture)
		end 

		if Layout.UseMana then 
			self.ExtraPower.Border:SetTexture(Layout.NoviceManaOrbTexture)
			self.ExtraPower.Border:SetVertexColor(unpack(Layout.NoviceManaOrbColor)) 
		end 

	end 

	
end 


-- Main Styling Function
local Style = function(self, unit, id, ...)

	-- Frame
	-----------------------------------------------------------
	self:SetSize(unpack(Layout.Size)) 
	self:Place(unpack(Layout.Place)) 

	if Layout.HitRectInsets then 
		self:SetHitRectInsets(unpack(Layout.HitRectInsets))
	else 
		self:SetHitRectInsets(0, 0, 0, 0)
	end 

	-- Assign our own global custom colors
	self.colors = Colors


	-- Scaffolds
	-----------------------------------------------------------

	-- frame to contain art backdrops, shadows, etc
	local backdrop = self:CreateFrame("Frame")
	backdrop:SetAllPoints()
	backdrop:SetFrameLevel(self:GetFrameLevel())
	
	-- frame to contain bars, icons, etc
	local content = self:CreateFrame("Frame")
	content:SetAllPoints()
	content:SetFrameLevel(self:GetFrameLevel() + 5)

	-- frame to contain art overlays, texts, etc
	local overlay = self:CreateFrame("Frame")
	overlay:SetAllPoints()
	overlay:SetFrameLevel(self:GetFrameLevel() + 10)

	-- Border
	-----------------------------------------------------------	
	if Layout.UseBorderBackdrop then 
		local border = self:CreateFrame("Frame")
		border:SetFrameLevel(self:GetFrameLevel() + 8)
		border:SetSize(unpack(Layout.BorderFrameSize))
		border:Place(unpack(Layout.BorderFramePlace))
		border:SetBackdrop(Layout.BorderFrameBackdrop)
		border:SetBackdropColor(unpack(Layout.BorderFrameBackdropColor))
		border:SetBackdropBorderColor(unpack(Layout.BorderFrameBackdropBorderColor))
		self.Border = border
	end 

	-- Health Bar
	-----------------------------------------------------------	
	local health 

	if (Layout.HealthType == "Orb") then 
		health = content:CreateOrb()

	elseif (Layout.HealthType == "SpinBar") then 
		health = content:CreateSpinBar()

	else 
		health = content:CreateStatusBar()
		health:SetOrientation(Layout.HealthBarOrientation or "RIGHT") 
		health:SetFlippedHorizontally(Layout.HealthBarSetFlippedHorizontally)
		if Layout.HealthBarSparkMap then 
			health:SetSparkMap(Layout.HealthBarSparkMap) -- set the map the spark follows along the bar.
		end 
	end 
	if (not Layout.UseProgressiveFrames) then 
		health:SetStatusBarTexture(Layout.HealthBarTexture)
		health:SetSize(unpack(Layout.HealthSize))
	end 

	health:Place(unpack(Layout.HealthPlace))
	health:SetSmoothingMode(Layout.HealthSmoothingMode or "bezier-fast-in-slow-out") -- set the smoothing mode.
	health:SetSmoothingFrequency(Layout.HealthSmoothingFrequency or .5) -- set the duration of the smoothing.
	health.colorTapped = Layout.HealthColorTapped  -- color tap denied units 
	health.colorDisconnected = Layout.HealthColorDisconnected -- color disconnected units
	health.colorClass = Layout.HealthColorClass -- color players by class 
	health.colorReaction = Layout.HealthColorReaction -- color NPCs by their reaction standing with us
	health.colorHealth = Layout.HealthColorHealth -- color anything else in the default health color
	health.frequent = Layout.HealthFrequentUpdates -- listen to frequent health events for more accurate updates
	self.Health = health
	
	if Layout.UseHealthBackdrop then 
		local healthBg = health:CreateTexture()
		healthBg:SetDrawLayer(unpack(Layout.HealthBackdropDrawLayer))
		healthBg:SetSize(unpack(Layout.HealthBackdropSize))
		healthBg:SetPoint(unpack(Layout.HealthBackdropPlace))
		if (not Layout.UseProgressiveFrames) then 
			healthBg:SetTexture(Layout.HealthBackdropTexture)
		end 
		self.Health.Bg = healthBg
	end 

	if Layout.UseHealthForeground then 
		local healthFg = health:CreateTexture()
		healthFg:SetDrawLayer("BORDER", 1)
		healthFg:SetSize(unpack(Layout.HealthForegroundSize))
		healthFg:SetPoint(unpack(Layout.HealthForegroundPlace))
		healthFg:SetTexture(Layout.HealthForegroundTexture)
		healthFg:SetDrawLayer(unpack(Layout.HealthForegroundDrawLayer))
		self.Health.Fg = healthFg
	end 

	-- Absorb Bar
	-----------------------------------------------------------	
	if Layout.UseAbsorbBar then 
		local absorb = content:CreateStatusBar()
		absorb:SetFrameLevel(health:GetFrameLevel() + 1)
		absorb:Place(unpack(Layout.AbsorbBarPlace))
		absorb:SetOrientation(Layout.AbsorbBarOrientation) -- grow the bar towards the left (grows from the end of the health)
		absorb:SetFlippedHorizontally(Layout.AbsorbBarSetFlippedHorizontally)

		if (not Layout.UseProgressiveFrames) then
			absorb:SetSize(unpack(Layout.AbsorbSize))
			absorb:SetStatusBarTexture(Layout.AbsorbBarTexture)
		end

		if Layout.AbsorbBarSparkMap then 
			absorb:SetSparkMap(Layout.AbsorbBarSparkMap) -- set the map the spark follows along the bar.
		end 

		absorb:SetStatusBarColor(unpack(Layout.AbsorbBarColor)) -- make the bar fairly transparent, it's just an overlay after all. 
		self.Absorb = absorb
	end 

	-- Power 
	-----------------------------------------------------------
	if Layout.UsePowerBar then 
		local power = backdrop:CreateStatusBar()
		power:SetSize(unpack(Layout.PowerSize))
		power:Place(unpack(Layout.PowerPlace))
		power:SetStatusBarTexture(Layout.PowerBarTexture)
		power:SetTexCoord(unpack(Layout.PowerBarTexCoord))
		power:SetOrientation(Layout.PowerBarOrientation or "RIGHT") -- set the bar to grow towards the top.
		power:SetSmoothingMode(Layout.PowerBarSmoothingMode) -- set the smoothing mode.
		power:SetSmoothingFrequency(Layout.PowerBarSmoothingFrequency or .5) -- set the duration of the smoothing.

		if Layout.PowerBarSparkMap then 
			power:SetSparkMap(Layout.PowerBarSparkMap) -- set the map the spark follows along the bar.
		end 

		power.ignoredResource = Layout.PowerIgnoredResource -- make the bar hide when MANA is the primary resource. 

		self.Power = power
		self.Power.OverrideColor = OverridePowerColor

		if Layout.UsePowerBackground then 
			local powerBg = power:CreateTexture()
			powerBg:SetDrawLayer(unpack(Layout.PowerBackgroundDrawLayer))
			powerBg:SetSize(unpack(Layout.PowerBackgroundSize))
			powerBg:SetPoint(unpack(Layout.PowerBackgroundPlace))
			powerBg:SetTexture(Layout.PowerBackgroundTexture)
			powerBg:SetVertexColor(unpack(Layout.PowerBackgroundColor)) 
			self.Power.Bg = powerBg
		end

		if Layout.UsePowerForeground then 
			local powerFg = power:CreateTexture()
			powerFg:SetSize(unpack(Layout.PowerForegroundSize))
			powerFg:SetPoint(unpack(Layout.PowerForegroundPlace))
			powerFg:SetDrawLayer(unpack(Layout.PowerForegroundDrawLayer))
			powerFg:SetTexture(Layout.PowerForegroundTexture)
			self.Power.Fg = powerFg
		end

		-- Power Value
		if Layout.UsePowerBar then 
			if Layout.UsePowerValue then 
				local powerVal = self.Power:CreateFontString()
				powerVal:SetPoint(unpack(Layout.PowerValuePlace))
				powerVal:SetDrawLayer(unpack(Layout.PowerValueDrawLayer))
				powerVal:SetJustifyH(Layout.PowerValueJustifyH)
				powerVal:SetJustifyV(Layout.PowerValueJustifyV)
				powerVal:SetFontObject(Layout.PowerValueFont)
				powerVal:SetTextColor(unpack(Layout.PowerValueColor))
				self.Power.Value = powerVal
				self.Power.UpdateValue = OverrideValue
			end 
		end		
	end 

	-- Cast Bar
	-----------------------------------------------------------
	if Layout.UseCastBar then
		local cast = content:CreateStatusBar()
		cast:SetSize(unpack(Layout.CastBarSize))
		cast:SetFrameLevel(health:GetFrameLevel() + 1)
		cast:Place(unpack(Layout.CastBarPlace))
		cast:SetOrientation(Layout.CastBarOrientation) -- set the bar to grow towards the right.
		cast:SetSmoothingMode(Layout.CastBarSmoothingMode) -- set the smoothing mode.
		cast:SetSmoothingFrequency(Layout.CastBarSmoothingFrequency)
		cast:SetStatusBarColor(unpack(Layout.CastBarColor)) -- the alpha won't be overwritten. 

		if (not Layout.UseProgressiveFrames) then 
			cast:SetStatusBarTexture(Layout.CastBarTexture)
		end 

		if Layout.CastBarSparkMap then 
			cast:SetSparkMap(Layout.CastBarSparkMap) -- set the map the spark follows along the bar.
		end

		self.Cast = cast
	end 


	-- Portrait
	-- TODO: fix!
	if Layout.UsePortrait then 
		local portrait = backdrop:CreateFrame("PlayerModel")
		portrait:SetPoint("TOPRIGHT", 73, 8)
		portrait:SetSize(85, 85) 
		portrait:SetAlpha(.85)
		portrait.distanceScale = 1
		portrait.positionX = 0
		portrait.positionY = 0
		portrait.positionZ = 0
		portrait.rotation = 0 -- in degrees
		portrait.showFallback2D = true -- display 2D portraits when unit is out of range of 3D models
		self.Portrait = portrait
		
		-- To allow the backdrop and overlay to remain 
		-- visible even with no visible player model, 
		-- we add them to our backdrop and overlay frames, 
		-- not to the portrait frame itself.  
		local portraitBg = backdrop:CreateTexture()
		portraitBg:SetDrawLayer("BACKGROUND", 0)
		portraitBg:SetPoint("TOPRIGHT", 116, 55)
		portraitBg:SetSize(173, 173)
		portraitBg:SetTexture(getPath("party_portrait_back"))
		portraitBg:SetVertexColor(.5, .5, .5) -- keep this dark
		self.Portrait.Bg = portraitBg

		local portraitShade = content:CreateTexture()
		portraitShade:SetDrawLayer("BACKGROUND", -1)
		portraitShade:SetPoint("TOPRIGHT", 83, 21)
		portraitShade:SetSize(107, 107) 
		portraitShade:SetTexture(getPath("shade_circle"))
		self.Portrait.Shade = portraitShade

		local portraitFg = content:CreateTexture()
		portraitFg:SetDrawLayer("BACKGROUND", 0)
		portraitFg:SetPoint("TOPRIGHT", 123, 61)
		portraitFg:SetSize(187, 187)
		self.Portrait.Fg = portraitFg
	end 


	-- Threat
	-----------------------------------------------------------	
	if Layout.UseThreat then 
		local threats = { IsShown = Threat_IsShown, Show = Threat_Show, Hide = Threat_Hide }
		threats.feedbackUnit = "player"
		threats.hideSolo = true
		threats.fadeOut = 3

		local threatHealth = backdrop:CreateTexture()
		threatHealth:SetDrawLayer("BACKGROUND", -2)
		threatHealth:SetTexCoord(1,0,0,1)
		threatHealth:SetAlpha(.75)
		threats.health = threatHealth

		local threatPortrait = backdrop:CreateTexture()
		threatPortrait:SetDrawLayer("BACKGROUND", -2)
		threatPortrait:SetAlpha(.75)
		threatPortrait:SetTexture(getPath("portrait_frame_glow"))
		threatPortrait:SetSize(187, 187)
		threatPortrait:SetPoint("CENTER", portraitFg, "CENTER", 0, 0)
		threats.portrait = threatPortrait

		self.Threat = threats
		self.Threat.OverrideColor = Threat_UpdateColor
	end

	-- Unit Level
	if Layout.UseLevel then 

		-- level text
		local level = overlay:CreateFontString()
		level:SetPoint("CENTER", self, "TOPRIGHT", 79, -62)
		level:SetDrawLayer("BORDER")
		level:SetJustifyH("CENTER")
		level:SetJustifyV("MIDDLE")
		level:SetFontObject(Fonts(12, true))
		level:SetShadowOffset(0, 0)
		level:SetShadowColor(0, 0, 0, 0)

		-- Hide the level of capped (or higher) players and NPcs 
		-- Doesn't affect high/unreadable level (??) creatures, as they will still get a skull.
		level.hideCapped = true 

		-- Hide the level of level 1's
		level.hideFloored = true

		-- Set the default level coloring when nothing special is happening
		level.defaultColor = { 251/255, 255/255, 255/255 } 
		level.alpha = .7

		-- Badge backdrop
		local levelBadge = overlay:CreateTexture()
		levelBadge:SetDrawLayer("BACKGROUND")
		levelBadge:SetPoint("CENTER", portrait, "CENTER", 48,-28)
		levelBadge:SetSize(86,86)
		levelBadge:SetTexture(getPath("point_plate"))
		levelBadge:SetVertexColor(Colors.ui.stone[1], Colors.ui.stone[2], Colors.ui.stone[3])
		level.Badge = levelBadge

		-- Skull texture for bosses, high level and dead units  
		local skull = overlay:CreateTexture()
		skull:SetDrawLayer("BORDER")
		skull:SetPoint("CENTER", levelBadge, "CENTER", 0, 0)
		skull:SetSize(64,64)
		skull:SetTexture(getPath("icon_skull"))
		level.Skull = skull

		-- Skull texture for bosses, high level and dead units  
		local dead = overlay:CreateTexture()
		dead:SetDrawLayer("BORDER")
		dead:SetPoint("CENTER", levelBadge, "CENTER", 0, 0)
		dead:SetSize(64,64)
		dead:SetTexture(getPath("icon_skull_dead"))
		dead:Hide()
		level.Dead = dead
		
		self.Level = level	
	end 

	-- Unit Classification (boss, elite, rare)
	if Layout.UseClassificationIndicator then 
		self.Classification = {}

		local boss = overlay:CreateTexture()
		boss:SetTexture(getPath("icon_classification_boss"))
		boss:SetSize(84,84)
		boss:SetPoint("CENTER", self, "BOTTOMRIGHT", 30, -1)
		boss:SetVertexColor(182/255, 183/255, 181/255)
		self.Classification.Boss = boss

		local elite = overlay:CreateTexture()
		elite:SetTexture(getPath("icon_classification_elite"))
		elite:SetSize(84,84)
		elite:SetPoint("CENTER", self, "BOTTOMRIGHT", 30, -1)
		elite:SetVertexColor(182/255, 183/255, 181/255)
		self.Classification.Elite = elite

		local rare = overlay:CreateTexture()
		rare:SetTexture(getPath("icon_classification_rare"))
		rare:SetSize(84,84)
		rare:SetPoint("CENTER", self, "BOTTOMRIGHT", 30, -1)
		rare:SetVertexColor(182/255, 183/255, 181/255)
		self.Classification.Rare = rare
	end


	-- Targeting
	-----------------------------------------------------------	
	-- Indicates who your target is targeting
	if Layout.UseTargetIndicator then 
		self.Targeted = {}

		local friend = overlay:CreateTexture()
		friend:SetTexture(getPath("icon_target_green"))
		friend:SetSize(96,48)
		friend:SetPoint("CENTER", self, "TOPRIGHT", -10, 12)
		friend:SetVertexColor(227/255, 231/255, 216/255)
		self.Targeted.YouByFriend = friend

		local enemy = overlay:CreateTexture()
		enemy:SetTexture(getPath("icon_target_red"))
		enemy:SetSize(96,48)
		enemy:SetPoint("CENTER", self, "TOPRIGHT", -10, 12)
		enemy:SetVertexColor(227/255, 231/255, 216/255)
		self.Targeted.YouByEnemy = enemy

		local pet = overlay:CreateTexture()
		pet:SetTexture(getPath("icon_target_blue"))
		pet:SetSize(96,48)
		pet:SetPoint("CENTER", self, "TOPRIGHT", -10, 12)
		pet:SetVertexColor(227/255, 231/255, 216/255)
		self.Targeted.PetByEnemy = pet
	end 

	-- Auras
	-----------------------------------------------------------
	if Layout.UseAuras then 
		local auras = content:CreateFrame("Frame")
		auras:Place(unpack(Layout.AuraFramePlace))
		auras:SetSize(unpack(Layout.AuraFrameSize)) -- auras will be aligned in the available space, this size gives us 8x1 auras
		auras.auraSize = Layout.AuraSize -- size of the aura. assuming squares. 
		auras.spacingH = Layout.AuraSpaceH -- horizontal/column spacing between buttons
		auras.spacingV = Layout.AuraSpaceV -- vertical/row spacing between aura buttons
		auras.growthX = Layout.AuraGrowthX -- auras grow to the left
		auras.growthY = Layout.AuraGrowthY -- rows grow downwards (we just have a single row, though)
		auras.maxVisible = Layout.AuraMax -- when set will limit the number of buttons regardless of space available
		auras.maxBuffs = Layout.AuraMaxBuffs -- maximum number of visible buffs
		auras.maxDebuffs = Layout.AuraMaxDebuffs -- maximum number of visible debuffs
		auras.debuffsFirst = Layout.AuraDebuffs -- show debuffs before buffs
		auras.showCooldownSpiral = Layout.ShowAuraCooldownSpirals -- don't show the spiral as a timer
		auras.showCooldownTime = Layout.ShowAuraCooldownTime -- show timer numbers
		auras.auraFilter = Layout.AuraFilter -- general aura filter, only used if the below aren't here
		auras.buffFilter = Layout.AuraBuffFilter -- buff specific filter passed to blizzard API calls
		auras.debuffFilter = Layout.AuraDebuffFilter -- debuff specific filter passed to blizzard API calls
		auras.AuraFilter = Layout.AuraFilterFunc -- general aura filter function, called when the below aren't there
		auras.BuffFilter = Layout.BuffFilterFunc -- buff specific filter function
		auras.DebuffFilter = Layout.DebuffFilterFunc -- debuff specific filter function
		auras.tooltipDefaultPosition = Layout.AuraTooltipDefaultPosition
		auras.tooltipPoint = Layout.AuraTooltipPoint
		auras.tooltipAnchor = Layout.AuraTooltipAnchor
		auras.tooltipRelPoint = Layout.AuraTooltipRelPoint
		auras.tooltipOffsetX = Layout.AuraTooltipOffsetX
		auras.tooltipOffsetY = Layout.AuraTooltipOffsetY
			
		self.Auras = auras
		self.Auras.PostCreateButton = PostCreateAuraButton -- post creation styling
		self.Auras.PostUpdateButton = PostUpdateAuraButton -- post updates when something changes (even timers)
	end 

	if Layout.UseBuffs then 
		local buffs = content:CreateFrame("Frame")
		buffs:Place(unpack(Layout.BuffFramePlace))
		buffs:SetSize(unpack(Layout.BuffFrameSize)) -- auras will be aligned in the available space, this size gives us 8x1 auras
		buffs.auraSize = Layout.BuffSize -- size of the aura. assuming squares. 
		buffs.spacingH = Layout.BuffSpaceH -- horizontal/column spacing between buttons
		buffs.spacingV = Layout.BuffSpaceV -- vertical/row spacing between aura buttons
		buffs.growthX = Layout.BuffGrowthX -- auras grow to the left
		buffs.growthY = Layout.BuffGrowthY -- rows grow downwards (we just have a single row, though)
		buffs.maxVisible = Layout.BuffMax -- when set will limit the number of buttons regardless of space available
		buffs.showCooldownSpiral = Layout.ShowBuffCooldownSpirals -- don't show the spiral as a timer
		buffs.showCooldownTime = Layout.ShowBuffCooldownTime -- show timer numbers
		buffs.debuffFilter = Layout.BuffFilter -- general aura filter, only used if the below aren't here
		buffs.BuffFilter = Layout.BuffFilterFunc -- general aura filter function, called when the below aren't there
		buffs.tooltipDefaultPosition = Layout.BuffTooltipDefaultPosition
		buffs.tooltipPoint = Layout.BuffTooltipPoint
		buffs.tooltipAnchor = Layout.BuffTooltipAnchor
		buffs.tooltipRelPoint = Layout.BuffTooltipRelPoint
		buffs.tooltipOffsetX = Layout.BuffTooltipOffsetX
		buffs.tooltipOffsetY = Layout.BuffTooltipOffsetY
			
		self.Buffs = buffs
		self.Buffs.PostCreateButton = PostCreateAuraButton -- post creation styling
		self.Buffs.PostUpdateButton = PostUpdateAuraButton -- post updates when something changes (even timers)
	end 

	if Layout.UseDebuffs then 
		local debuffs = content:CreateFrame("Frame")
		debuffs:Place(unpack(Layout.DebuffFramePlace))
		debuffs:SetSize(unpack(Layout.DebuffFrameSize)) -- auras will be aligned in the available space, this size gives us 8x1 auras
		debuffs.auraSize = Layout.DebuffSize -- size of the aura. assuming squares. 
		debuffs.spacingH = Layout.DebuffSpaceH -- horizontal/column spacing between buttons
		debuffs.spacingV = Layout.DebuffSpaceV -- vertical/row spacing between aura buttons
		debuffs.growthX = Layout.DebuffGrowthX -- auras grow to the left
		debuffs.growthY = Layout.DebuffGrowthY -- rows grow downwards (we just have a single row, though)
		debuffs.maxVisible = Layout.DebuffMax -- when set will limit the number of buttons regardless of space available
		debuffs.showCooldownSpiral = Layout.ShowDebuffCooldownSpirals -- don't show the spiral as a timer
		debuffs.showCooldownTime = Layout.ShowDebuffCooldownTime -- show timer numbers
		debuffs.debuffFilter = Layout.DebuffFilter -- general aura filter, only used if the below aren't here
		debuffs.DebuffFilter = Layout.DebuffFilterFunc -- general aura filter function, called when the below aren't there
		debuffs.tooltipDefaultPosition = Layout.DebuffTooltipDefaultPosition
		debuffs.tooltipPoint = Layout.DebuffTooltipPoint
		debuffs.tooltipAnchor = Layout.DebuffTooltipAnchor
		debuffs.tooltipRelPoint = Layout.DebuffTooltipRelPoint
		debuffs.tooltipOffsetX = Layout.DebuffTooltipOffsetX
		debuffs.tooltipOffsetY = Layout.DebuffTooltipOffsetY
			
		self.Debuffs = debuffs
		self.Debuffs.PostCreateButton = PostCreateAuraButton -- post creation styling
		self.Debuffs.PostUpdateButton = PostUpdateAuraButton -- post updates when something changes (even timers)
	end 

	-- Texts
	-----------------------------------------------------------	

	-- Unit Name
	if Layout.UseName then 
		local name = overlay:CreateFontString()
		name:SetPoint(unpack(Layout.NamePlace))
		name:SetDrawLayer(unpack(Layout.NameDrawLayer))
		name:SetJustifyH(Layout.NameDrawJustifyH)
		name:SetJustifyV(Layout.NameDrawJustifyV)
		name:SetFontObject(Layout.NameFont)
		name:SetTextColor(unpack(Layout.NameColor))
		if Layout.NameSize then 
			name:SetSize(unpack(Layout.NameSize))
		end 
		self.Name = name
	end 

	-- Health Value
	if Layout.UseHealthValue then 
		local healthVal = health:CreateFontString()
		healthVal:SetPoint(unpack(Layout.HealthValuePlace))
		healthVal:SetDrawLayer(unpack(Layout.HealthValueDrawLayer))
		healthVal:SetJustifyH(Layout.HealthValueJustifyH)
		healthVal:SetJustifyV(Layout.HealthValueJustifyV)
		healthVal:SetFontObject(Layout.HealthValueFont)
		healthVal:SetTextColor(unpack(Layout.HealthValueColor))

		-- Health percentage for bosses
		-- TODO: FIX!
		if Layout.UseHealthPercent then 
			local healthPerc = health:CreateFontString()
			healthPerc:SetPoint("LEFT", 27, 4)
			healthPerc:SetDrawLayer("OVERLAY")
			healthPerc:SetJustifyH("CENTER")
			healthPerc:SetJustifyV("MIDDLE")
			healthPerc:SetFontObject(Fonts(18, true))
			healthPerc:SetShadowOffset(0, 0)
			healthPerc:SetShadowColor(0, 0, 0, 0)
			healthPerc:SetTextColor(240/255, 240/255, 240/255, .5)
		end 
		
		self.Health.Value = healthVal
		self.Health.Percent = healthPerc
		self.Health.OverrideValue = Layout.HealthValueOverride ~= nil and Layout.HealthValueOverride or OverrideHealthValue
	end 

	-- Absorb Value
	if Layout.UseAbsorbBar then 
		if Layout.UseAbsorbValue then 
			local absorbVal = overlay:CreateFontString()
			if Layout.AbsorbValuePlaceFunction then 
				absorbVal:SetPoint(Layout.AbsorbValuePlaceFunction(self))
			else 
				absorbVal:SetPoint(unpack(Layout.AbsorbValuePlace))
			end 
			absorbVal:SetDrawLayer(unpack(Layout.AbsorbValueDrawLayer))
			absorbVal:SetJustifyH(Layout.AbsorbValueJustifyH)
			absorbVal:SetJustifyV(Layout.AbsorbValueJustifyV)
			absorbVal:SetFontObject(Layout.AbsorbValueFont)
			absorbVal:SetTextColor(unpack(Layout.AbsorbValueColor))
			self.Absorb.Value = absorbVal 
			self.Absorb.OverrideValue = OverrideValue
		end 
	end 

	-- Update textures according to player level
	if Layout.UseProgressiveFrames then 
		PostUpdateTextures(self)
	end 
end

Module.GetFrame = function(self)
	return self.frame
end 

Module.OnEvent = function(self, event, ...)
	if (event == "PLAYER_TARGET_CHANGED") then
	
		if UnitExists("target") then
			-- Play a fitting sound depending on what kind of target we gained
			if UnitIsEnemy("target", "player") then
				self:PlaySoundKitID(SOUNDKIT.IG_CREATURE_AGGRO_SELECT, "SFX")
			elseif UnitIsFriend("player", "target") then
				self:PlaySoundKitID(SOUNDKIT.IG_CHARACTER_NPC_SELECT, "SFX")
			else
				self:PlaySoundKitID(SOUNDKIT.IG_CREATURE_NEUTRAL_SELECT, "SFX")
			end

			-- Update textures according to player level
			if Layout.UseProgressiveFrames then 
				PostUpdateTextures(self:GetFrame())
			end 
		else
			-- Play a sound indicating we lost our target
			self:PlaySoundKitID(SOUNDKIT.INTERFACE_SOUND_LOST_TARGET_UNIT, "SFX")
		end

	elseif (event == "PLAYER_LEVEL_UP") then 
		local level = ...
		if (level and (level ~= LEVEL)) then
			LEVEL = level
		else
			local level = UnitLevel("player")
			if (level ~= LEVEL) then
				LEVEL = level
			end
		end
	end
end

Module.OnInit = function(self)
	local targetFrame = self:SpawnUnitFrame("target", "UICenter", Style)
	self.frame = targetFrame
end 

Module.OnEnable = function(self)
	self:RegisterEvent("PLAYER_LEVEL_UP", "OnEvent")
	self:RegisterEvent("PLAYER_TARGET_CHANGED", "OnEvent")
end 
